""" 
# Hdl21 PDK Interface

Hdl21 "process development kit" (PDK) packages provide the bundle between the Python HDL and fabricatable process technologies. 
PDK packages typically include: 

* (a) Definitions of technology-specific `ExternalModule`s, commonly including transistors and passive components, and 
* (b) A transformation method for converting generic `Primitive` elements to the technology-specific `ExternalModule`s.

The latter manifests as a "compiler pass" over the Hdl21 circuit-proto tree. 
An Hdl21 design hierarchically traversed and instances of `hdl21.Primitive`s 
are transformed, typically to technology-specific `ExternalModule`s. 

The PDK API consists of one required method `compile`: 

```python
def compile(src: Elaboratables) -> None
```

The core method `compile` transforms a process-generic design hierarchy into PDK-specific content. 
This will commonly manifest as replacement of `hdl21.Primitive` instances with PDK-specific `ExternalModule`s. 

The `Elaboratables` union type which serves as the input of `compile` can be any of: 

* An `hdl21.Module`
* A call to an `hdl21.Generator`
* Lists thereof

A typical implementation of `compile`, shown in the `hdl21.pdk.sample_pdk` package, is to call the base class `visit_elaboratables` 
method on a `Walker` subclass which implements the desired PDK transformations. For example: 

```python
def compile(src: h.Elaboratables) -> None:
    # Compile `src` to the Sample technology 
    return SamplePdkWalker().visit_elaboratables(src)
```

## Plug-in System 

Plug-in-style registration of `hdl21.PDK`s makes them available to `hdl21.Generator`s, 
and generally allows for a full program-worth of comprehension of the target process.  
Registering a PDK requires only passing its (Python) module as an argument to the `hdl21.pdk.register` function. 
Typical PDKs packages then take the typical form: 

```python
from hdl21.pdk import register

# Grab our primary PDK-definition module
from . import sample_pdk

# And register as a PDK module
register(sample_pdk)
```

```python
# sample_pdk/sample_pdk.py

import hdl21 as h

# Definitions of PDK modules 
MyMos = h.ExternalModule(name='MyMos', ...) 

# Compilation method 
def compile(src: h.Elaboratables) -> h.Elaboratables:
    ...

"""

import inspect
from typing import Optional, Union
from types import ModuleType

from ..elab import Elaboratables


class _PdkManager:
    """The "private" singleton manager of available PDK modules"""

    _the_one = None

    def __new__(cls, *_, **__):
        if cls._the_one is None:
            cls._the_one = super().__new__(cls)
        return cls._the_one

    def __init__(self):
        self.modules = set()
        self.names = dict()
        self.default = None


# Create that singleton
_mgr = _PdkManager()


"""
## Public API Methods 
"""


def register(module: ModuleType) -> None:
    """Register the (Python) Module `module` as a PDK"""
    if module in _mgr.modules:
        return

    # Checks on the PDK `module`
    if not isinstance(module, ModuleType):
        raise TypeError(f"{module} is not a Python Module")
    if not hasattr(module, "compile"):
        raise TypeError("PDK module must have a compile() method")

    # Do a buncha checking on this method here
    sig = inspect.signature(module.compile)

    args = list(sig.parameters.values())
    if len(args) != 1:
        raise RuntimeError(f"Invalid call signature for {module}.compile: {args}")

    # Extract the parameters-argument type
    paramtype = args[0].annotation
    if paramtype is not Elaboratables:
        msg = f"Invalid call signature for {module}.compile. Argument type must be `hdl21.Elaboratables`, not {paramtype}"
        raise RuntimeError(msg)

    # Validate the return type is also `Package`
    rt = sig.return_annotation
    if rt not in (None, type(None)):
        msg = f"Invalid call signature for {module}.compile. Return type must be `None`, not {paramtype}"
        raise RuntimeError(msg)

    # Checks out. Add it.
    _mgr.modules.add(module)
    _mgr.names[module.__name__] = module


def compile(
    src: Elaboratables, pdk: Optional[Union[str, ModuleType]] = None
) -> Elaboratables:
    """
    Compile to a target PDK.

    Uses the optional `pdk` argument as a target, if provided and valid.
    Otherwise uses the default PDK module.
    Raises a `RuntimeError` if there is no unambiguous default.
    """
    if pdk is None:
        pdk = default()
    elif isinstance(pdk, str):
        # Grab by-name from our registered names-dict
        pdk = _mgr.names.get(pdk, None)
        if pdk is None:
            msg = f"No PDK named {pdk}"
            raise RuntimeError(msg)
    elif isinstance(pdk, ModuleType):
        # Ensure that `pdk` is registered and checked as a valid PDK module
        _mgr.register(pdk)

    if pdk is None:  # Check for no-default-available cases
        if not len(_mgr.modules):
            raise RuntimeError("No PDK modules registered")

        msg = f"Multiple ({len(_mgr.modules)}) PDK modules registered: [\n"
        for m in _mgr.modules:
            msg += "\t" + str(m) + "\n"
        msg += "] \n"
        msg += f"Set one as the default via `hdl21.pdk.set_default()` (or remove all others) to use `h.pdk.compile()`."
        raise RuntimeError(msg)

    # Run the compiler
    return pdk.compile(src)


def set_default(to: Union[ModuleType, str]) -> None:
    """Set the default PDK to use when no PDK is specified in a `hdl21.Generator`"""

    if isinstance(to, str):
        to = _mgr.names.get(to, None)
        if to is None:
            raise RuntimeError(f"No PDK named {to} registered")
    elif isinstance(to, ModuleType):
        if to not in _mgr.modules:
            raise RuntimeError(f"No PDK named {to} registered")
    else:
        raise TypeError

    # Checks out; set as our default
    _mgr.default = to


def default() -> Optional[ModuleType]:
    """Retrieve the default PDK plug-in, if there is an unambiguous one set up."""

    if _mgr.default is not None:
        return _mgr.default
    if len(_mgr.modules) == 1:
        # Get the only element from the set
        return next(iter(_mgr.modules))
    return None



"""
# Mark each Module as elaborated
"""

# Local imports
from ...module import Module
from .base import ElabPass


class MarkModules(ElabPass):
    def elaborate_module(self, module: Module) -> Module:
        # Check that every module has a name
        if not module.name:
            msg = f"Anonymous Module {module} cannot be elaborated (did you forget to name it?)"
            self.fail(msg)

        # `_elaborated` is the module itself,
        # and not just a boolean, in case we want to
        # have differences between the two some day.
        module._elaborated = module

        return module



"""
# No-Connect

Special placeholder connectable-object which indicates "unconnected" Ports,
typically unconnected outputs.

An optional `name` field allows guidance for external netlisting,
for cases in which consistent naming is desirable (e.g. for waveform probing).
"""

from typing import Optional, Set

# Local imports
from .datatype import datatype
from .connect import connectable
from .concat import concatable
from .portref import PortRef


@concatable
@connectable
@datatype
class NoConn:
    """
    # No-Connect

    Special placeholder connectable-object which indicates "unconnected" Ports,
    typically unconnected outputs.

    An optional `name` field allows guidance for external netlisting,
    for cases in which consistent naming is desirable (e.g. for waveform probing).
    """

    name: Optional[str] = None

    def __post_init__(self) -> None:
        # Internal management data
        # Connected port references
        self._connected_ports: Set[PortRef] = set()

    def __eq__(self, other: "NoConn") -> bool:
        """`NoConn`s are "equal" only if identical objects."""
        return self is other

    def __hash__(self) -> int:
        """`NoConn`s are "equal" only if identical objects."""
        return hash(id(self))



"""
# PDK Installations 

In addition to their in-memory, Hdl21-representable content, 
process development kits (PDK)s often require large suites of files 
which represent the underlying technology. 

Transistor models and SPICE "library" files are common examples 
pertinent to Hdl21. Tech-files, layout libraries, and the like 
are similarly necessary for related pieces of software. 

Each `hdl21.pdk.PdkInstallation` represents a typical *site installation*, 
i.e. the combination of that content and its location on a 
particular file system. 

Tracking such site-specific content through long call-stacks 
can be pedantic, especially since the information is generally 
shared throughout a python process. The typical usage to avoid this is to, 
in a "site-specific" module or package, 

* Import the target PDK module 
* Create an instance of its `PdkInstallation` subtype
* Affix that instance to the PDK module's `install` member 

E.g. 
```python
# "sitepdks.py", or similar

import mypdk 

mypdk.install = mypdk.Install(
    models = "/path/to/models",
    path2 = "/path/2",
    # etc.
)
```

Any programs needing the site-specific data can then import `mysite`, 
while code directly requiring the contents of the `PdkInstall` need 
not be directly aware of its contents. 
For example when invoking a simulation requiring `mypdk`'s models:

```python 
# Simulation Invocation 
# Presumes `mysite` or similar has set `mypdk.install` beforehand. 

import mypdk
sim = Sim(tb=tb)
sim.lib(mypdk.install.models, "ss")
sim.run()
```

"""

from pydantic.dataclasses import dataclass


@dataclass
class PdkInstallation:
    """
    # PDK Installation

    A named set of PDK-related data, typically including details of
    a particular "site installation" of external files and settings.

    While often comprised of file-system `Path`s, the content of a `PdkInstallation`
    can in principle be anything related to the installation,
    e.g. compilation settings for a particular tool, or metadata indicating
    relevant tool versions.

    Most PDK modules should generally create a `pydantic.dataclass` which is also
    a sub-class of `PdkInstallation`, e.g.

    ```python
    # mypdk.py

    from pathlib import Path
    from pydantic.dataclasses import dataclass
    from hdl21.pdk import PdkInstallation

    @dataclass
    class Install(PdkInstallation):
        model_file: Path
        options: Dict[str, str]
    ```

    ```python
    # site.py
    import mypdk

    mypdk.install = mypdk.Install(
        model_file = "/my/pdk/models",
        options = {"opt" : "max"}
    )
    ```

    ```python
    # Application-Level Code
    # Running an example simulation with `mypdk`

    import site   # Creates the site-specific `PdkInstallation`s
    import mypdk  # Import the target PDK

    sim = Sim(tb=tb)
    sim.lib(mypdk.install.models, "ss")
    sim.run()
    ```

    This "installation data" can in principle be stored anywhere, or in unstructured types.
    The primary goal of the `PdkInstallation` type is to centralize this data,
    and to enable shorthand registration and recall, particularly for
    the common case of a single in-memory `PdkInstallation`.
    """

    ...  # No content



"""
# Hdl21 Attribute Magic 

Facilities for all of the `getattr` and `setattr` "magic" used by Hdl21 types. 

There is fairly little "magic" (ahem) being performed here; 
this module serves as more of a written reminder of how Hdl21's `getattr` and `setattr` overrides work. 

Getting them completely straight can be a pain, including a few complications: 
* During object initialization, and for setting "built-ins" e.g. the "dunder" methods 
* Many failures tend to occur only while *debugging*, e.g. with `pdb` or an IDE plugin. 
  * This makes tracking down problems with this scheme aways more difficult, as unit-tests fail to catch them. 

The scheme adopted here is what we've empirically observed to be the most reliable. 
* A `_initialized` boolean field is set to `False` as the VERY FIRST thing in object construction.
* All "magic methods" check the value of this field, and run "regular" `setattr` and `getattr` if it is `False`.
* The `_initialized` field is then set to `True` at the end of object construction.

The methods here primarily serve as a central reminder of and reusable source for this scheme. 
"""

from typing import TypeVar, Type, Any

T = TypeVar("T")


def init(cls: Type[T]) -> Type[T]:
    """
    Initialization decorator for "attr magic" types.

    Sets an `_initialized` field to `False` as part of a `__new__` method, so that `__init__` need not remember to do so.
    Constructors are required to set `_initialized` to `True` when *all* construction has completed.
    For inherited types, this generally includes both parent and subclass initialization.

    NOTE: this is a very important part of the `setattr` and `getattr` methods,
    particularly for their "bootstrapping phase" during object creation.
    """

    if "__new__" in cls.__dict__:
        msg = f"Internal Error: `__new__` doubly defined for {cls}"
        raise RuntimeError(msg)

    def __new__(cls, *_, **__):
        inst = object.__new__(cls)
        object.__setattr__(inst, "_initialized", False)
        return inst

    cls.__new__ = __new__
    return cls


def no_setattr(cls: Type[T]) -> Type[T]:
    """
    Decorator to "freeze" an object's attributes after construction.
    This is not called "frozen" or similar (unlike the std-lib),
    because it is much more difficult to actually freeze a Python object's inner fields.
    So, `no_setattr` types *are not* immutable - their attributes can still be changed,
    but *new attributes* cannot be added.
    """

    if "__setattr__" in cls.__dict__:
        msg = f"Internal Error: `__setattr__` doubly defined for {cls}"
        raise RuntimeError(msg)

    def __setattr__(self, key: str, val: Any) -> None:
        """Only allow setting special attributes after initialization."""

        special: bool = key.startswith("_")  # Special variables are always allowed
        inited: bool = object.__getattribute__(self, "_initialized")

        if special or not inited:
            return object.__setattr__(self, key, val)

        # FIXME: maybe allow customizing the error handler?
        msg = f"Cannot set attributes on {self} - attempting {key} = {val}"
        raise RuntimeError(msg)

    cls.__setattr__ = __setattr__
    return cls


def only_set_known_attrs(cls: Type[T]) -> Type[T]:
    """
    Decorator for "only allow setting existing attributes".
    Allows setting attributes only under three conditions:
    * The attribute name starts with an underscore, e.g. `__init__`, or
    * The object is still in its "bootstrapping phase", i.e. `_initialized` is `False`, or
    * The attribute already exists on the object
    """

    if "__setattr__" in cls.__dict__:
        msg = f"Internal Error: `__setattr__` doubly defined for {cls}"
        raise RuntimeError(msg)

    def __setattr__(self, key: str, val: Any) -> None:
        """Only allow setting existing attributes."""

        special: bool = key.startswith("_")  # Special variables are always allowed
        inited: bool = object.__getattribute__(self, "_initialized")

        if special or not inited:
            return object.__setattr__(self, key, val)

        # Check if the attribute already exists
        # This generally wants to "reach around" any of our other `__getattr__` magic.
        # If it doesn't exist, we'll get an exception, which we'll catch and re-raise.
        try:
            object.__getattribute__(self, key)
        except:
            # FIXME: maybe allow customizing the error handler?
            msg = f"Cannot set attributes on {self} - attempting {key} = {val}"
            raise RuntimeError(msg)
        else:
            return object.__setattr__(self, key, val)

    cls.__setattr__ = __setattr__
    return cls



""" 
# Hdl21 Parameters 
# Unit Tests 
"""

import pytest
from dataclasses import field, FrozenInstanceError

from pydantic import ValidationError
from pydantic.dataclasses import dataclass

# Import the PUT (package under test)
import hdl21 as h


def test_params1():
    # Initial param-class test

    @h.paramclass
    class MyParams:
        a = h.Param(dtype=int, default=5, desc="your fave")

    assert h.isparamclass(MyParams)

    m = MyParams()
    assert isinstance(m.a, int)
    assert m.a == 5
    assert isinstance(m.__params__["a"], h.Param)
    assert m.defaults() == dict(a=5)
    assert m.descriptions() == dict(a="your fave")


def test_params2():
    @h.paramclass
    class Pc2:
        # Required
        r = h.Param(dtype=str, desc="required")
        # Optional
        o = h.Param(dtype=str, default="hmmm", desc="optional")

    assert h.isparamclass(Pc2)
    assert Pc2.defaults() == dict(o="hmmm")
    assert Pc2.descriptions() == dict(r="required", o="optional")

    p = Pc2(r="provided")
    assert isinstance(p.r, str)
    assert p.r == "provided"
    assert isinstance(p.o, str)
    assert p.o == "hmmm"
    assert p.defaults() == dict(o="hmmm")
    assert p.descriptions() == dict(r="required", o="optional")


def test_params3():
    # Test parameters with a creation-time list-tuple conversion
    @h.paramclass
    class HasTuple:
        t = h.Param(dtype=tuple, desc="Go ahead, try a list")

    ht = HasTuple(t=[1, 2, 3])
    assert isinstance(ht.t, tuple)
    assert ht.t == (1, 2, 3)


def test_params4():
    # Test some param-class nesting
    from dataclasses import asdict

    @h.paramclass
    class Inner:
        i = h.Param(dtype=int, desc="Inner int-field")

    @h.paramclass
    class Outer:
        inner = h.Param(dtype=Inner, desc="Inner fields")
        f = h.Param(dtype=float, desc="A float", default=3.14159)

    o = Outer(inner=Inner(11))
    assert isinstance(o, Outer)
    assert isinstance(o.inner, Inner)
    assert o.inner == Inner(11)
    assert o.inner.i == 11
    assert o.f == 3.14159

    # Create from a (nested) dictionary
    d1 = {"inner": {"i": 11}, "f": 22.2}
    o1 = Outer(**d1)
    uname1 = h.params._unique_name(o1)
    # Convert back to another dictionary
    d2 = asdict(o1)
    # And check they line up
    assert d1 == d2
    # Round-trip back to an `Outer`
    o2 = Outer(**d2)
    uname2 = h.params._unique_name(o2)
    assert uname1 == uname2
    assert uname1 == "3dcc309796996b3a8a61db66631c5a93"


def test_bad_params1():
    # Test a handful of errors Params and paramclasses should raise.

    with pytest.raises(RuntimeError):
        # Test that creating a h.paramclass with parent-class(es) fails

        @h.paramclass
        class C(TabError):  # Of course this is a sub-class of the best built-in class
            ...

    @h.paramclass
    class C:
        a = h.Param(dtype=int, desc="Gonna Fail!")

    with pytest.raises(RuntimeError):
        # Test that sub-classing a h.paramclass fails

        class D(C):
            ...

    with pytest.raises((TypeError, ValidationError)):
        # Test that missing arguments fail
        # Note whether this is `TypeError` or `ValidationError` depends on the version of Pydantic.
        c = C()

    with pytest.raises(ValidationError):
        # Test invalid argument types fail

        c = C(a=TabError)

    with pytest.raises(FrozenInstanceError):
        # Test that attempts at mutating h.paramclasses fail
        c = C(a=3)
        c.a = 4

    with pytest.raises(RuntimeError):
        # Test "no Module sub-classing"

        class E(h.Module):
            ...

    with pytest.raises(RuntimeError):
        # Test "no decorating inherited types"

        @h.module
        class E2(TabError):
            ...

    with pytest.raises(RuntimeError):
        # Test bad parameter names

        @h.paramclass
        class P:
            descriptions = h.Param(dtype=str, desc="", default="BAD!")

    with pytest.raises(RuntimeError):
        # Test bad parameter names

        @h.paramclass
        class P:
            defaults = h.Param(dtype=str, desc="", default="BAD!")

    with pytest.raises(RuntimeError):
        # Test non-params in `@paramclass`

        @h.paramclass
        class P:
            something = 11

    with pytest.raises(RuntimeError):
        # Test a bad argument type
        h.params._unique_name(33)


def test_param_default_factory():
    """Test the `default_factory` feature of `Param`"""

    # Test the pydantic versions without `paramclass` first

    @dataclass
    class NoFactory:
        i: int = 11

    @dataclass
    class HasFactory:
        a: int = field(default_factory=lambda: 5)
        l: list = field(default_factory=list)
        no: NoFactory = field(default_factory=NoFactory)

    has = HasFactory()

    assert has.a == 5
    assert has.l == []
    assert has.no == NoFactory(i=11)

    # And test `paramclass`es

    @h.paramclass
    class NoFactory:
        i = h.Param(dtype=int, desc="i", default=11)

    @h.paramclass
    class HasFactory:
        a = h.Param(dtype=int, desc="a", default_factory=lambda: 5)
        l = h.Param(dtype=list, desc="l", default_factory=list)
        no = h.Param(dtype=NoFactory, desc="no", default_factory=NoFactory)

    has = HasFactory()

    assert has.a == 5
    assert has.l == []
    assert has.no == NoFactory(i=11)



"""
# Helper for resolving the "types" of references. 

Note "types" here does not mean Python-types, but something more like "connection types". 
Resolving a reference "type" produces either a Signal or BundleInstance, 
which is "compatible with", but not equal to, the ultimate referent. 

For example: 

```python
@h.module
class Inner:
    p = h.Port()

@h.module
class Outer:
    i = Inner() 

# Create a `PortRef`
ref = Outer.i.p

# All of these succeed:
assert ref is Outer.i.p
assert resolve_portref_type(ref) is not Outer.i.p
assert resolve_portref_type(ref) is not Inner.p
```

Here `ref` will ultimately refer to the a Signal within an `Instance` of `Inner`. 
At the time of calling `resolve_portref_type`, no such Signal exists. 
The result of `resolve_portref_type` is therefore the `Inner` *Module definition level* Signal object. 
This is often required e.g. for connection-validity checking, or for copying to new Signals.

The important thing is: don't take the results of `resolve_*_type` as the actual referents! 
"""

# Std-Lib Imports
from typing import Union, Callable

# Local imports
from ...portref import PortRef
from ...connect import Connectable
from ...signal import Signal
from ...bundle import (
    BundleInstance,
    BundleRef,
)


def fail(msg: str) -> None:
    # The default failure handler
    raise RuntimeError(msg)


def resolve_bundleref_type(
    bref: BundleRef, failer: Callable = fail
) -> Union[Signal, BundleInstance]:
    """
    Resolve a bundle-reference to either a `Signal` or sub-`Bundle` Instance.

    NOTE this returns a *representative* signal or bundle instance,
    i.e. one with the correct type and width - not *the signal* for a given instance.
    In other words: this is fine for connection-validity checking,
    but *not* for copying signals during flattening.
    Hence the "type" name suffix, although what we return is not really a type.
    """

    if isinstance(bref.parent, BundleInstance):  # Parent is a BundleInstance.
        parent = bref.parent
    elif isinstance(
        bref.parent, BundleRef
    ):  # Nested reference. Recursively resolve the parent.
        parent = resolve_bundleref_type(bref.parent, failer)
    else:
        failer(f"Invalid BundleRef parent for {bref}")

    # Get the attribute from the parent namespace, or fail if not available.
    attr = parent.of.get(bref.attrname)
    if attr is None:
        msg = f"Bundle `{bref.parent.of.name}` has no attribute `{bref.attrname}`"
        failer(msg)
    return attr


def resolve_portref_type(
    pref: PortRef, failer: Callable = fail
) -> Union[Signal, BundleInstance]:
    """Get the `Port` object referred to by `pref`."""

    instantiable = pref.inst.of

    if pref.portname in instantiable.ports:
        return instantiable.ports[pref.portname]

    if (
        hasattr(instantiable, "bundle_ports")
        and pref.portname in instantiable.bundle_ports
    ):
        return instantiable.bundle_ports[pref.portname]

    return failer(f"Invalid PortRef {pref}")


def update_ref_deps(ref: Union[PortRef, BundleRef], resolved: Connectable):
    """Update all downstream dependencies on a `Ref` after it has been resolved to `resolved`."""

    # Reconnect all connected ports
    for connected_port in list(ref._connected_ports):
        connected_port.inst.replace(connected_port.portname, resolved)

    # Update all dependent slices and concats
    if hasattr(ref, "_slices"):
        for slice_ in ref._slices:
            slice_.parent = resolved
    if hasattr(ref, "_concats"):
        for concat in ref._concats:
            parts = list(concat.parts)
            parts = [resolved if p is ref else p for p in parts]
            concat.parts = tuple(parts)



"""
# Slice-Enabling Decorator
"""

from typing import Union, TypeVar, Type

from .connect import is_connectable

T = TypeVar("T")


def sliceable(cls: Type[T]) -> Type[T]:
    """Decorator to add the 'square-bracket indexing produces `Slice`s' behavior."""

    if getattr(cls, "__getitem__", None) is not None:
        msg = f"Internal hdl21 Error: invavlid `slices`-decoration of {cls} with existing __getitem__"
        raise RuntimeError(msg)

    if not is_connectable(cls):
        msg = f"Internal hdl21 Error: invavlid `slices`-decoration of {cls} which is not `connectable`"
        raise RuntimeError(msg)

    def __getitem__(self, index: Union[int, slice]) -> "Slice":
        return _slice(parent=self, index=index)

    # Add the new behavior to the class
    cls.__getitem__ = __getitem__
    cls.__getitem__.__doc__ = _slice.__doc__
    # And a marker attribute
    cls.__slices__ = True
    # And don't forget to return that class!
    return cls


def is_sliceable(obj: object) -> bool:
    """Returns True if the class has the `slices` decorator."""
    return getattr(obj, "__slices__", False)


def _slice(*, parent: "Sliceable", index: Union[int, slice]) -> "Slice":
    """
    Square-bracket slicing into Signals, Concatenations, and other Slices.
    Assuming valid inputs, returns a signal-`Slice`.

    Hdl21 slices are indexed "Python style", in the senses that:
    * Negative indices are supported, and count from the "end" of the Signal.
    * Slice-ranges such as `sig[0:2]` are supported, and *inclusive* of the start, while *exclusive* of the end index.
    * Negative-range slices such as `sig[2:0:-1]`, again *inclusive* of the start, *exclusive* of the end index, and *reversed*.

    Popular HDLs commonly use different signal-indexing conventions.
    Hdl21's own primary exchange format (in ProtoBuf) does as well,
    eschewing adopting inclusive-endpoints and negative-indexing.
    """

    from .slice import Slice

    if not isinstance(index, (int, slice)):
        raise TypeError

    slize = Slice(parent=parent, index=index)
    parent._slices.add(slize)
    return slize



"""
# Metric Prefixes

Defines the `Prefix` enumeration of SI unit-prefixes, 
and the `Prefixed` combination of a `Prefix` and a literal value. 

Common prefixes such as µ (micro), n (nano), and K (kilo) 
are also exposed as single-character identifiers.
Most commonly these can be used with the multiplication operator 
in expressions such as `5 * n`, `11 * M`, and `1 * µ`.

For generation of other prefixes an `e()` function allows for syntax similar to 
typical floating-point value generation. This is again most commonly useful in conjunction 
with the multiplication operator, to construct values such as `11 * e(-21)`. 

"""

from enum import Enum
from decimal import Decimal
from typing import Optional, Any, Union, Tuple
from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass


EPSILON = 20


class Prefix(Enum):
    """Enumerated Unit Prefixes
    Values are equal to the associated power-of-ten exponent."""

    YOCTO = -24
    ZEPTO = -21
    ATTO = -18
    FEMTO = -15
    PICO = -12
    NANO = -9
    MICRO = -6
    MILLI = -3
    CENTI = -2
    DECI = -1
    UNIT = 0  # No prefix
    DECA = 1
    HECTO = 2
    KILO = 3
    MEGA = 6
    GIGA = 9
    TERA = 12
    PETA = 15
    EXA = 18
    ZETTA = 21
    YOTTA = 24

    @classmethod
    def from_exp(cls, exp: int) -> Optional["Prefix"]:
        """Get the prefix from the exponent. If `exp` is not
        in the members of Prefix, returns None instead"""
        inverted = {v.value: v for v in cls.__members__.values()}

        return inverted.get(exp, None)

    @classmethod
    def closest(cls, exp: Any) -> Optional["Prefix"]:
        return min(cls.__members__.values(), key=lambda x: abs(x.value - exp))

    def __mul__(self, other: Any):

        if isinstance(other, Prefix):
            # Prefix times Prefix, eg. `p * M == µ`
            targ = self.value + other.value
            return e(targ)
        elif isinstance(other, (str, int, float, Decimal, Prefixed)):
            # Prefix times number, eg. `p * 5 == 5 * p == 5 * Prefix.UNIT`
            return self.__rmul__(other)

        return NotImplemented

    def __rmul__(self, other: Any):
        """Right-hand-side multiplication operator, e.g. `5 * µ`."""

        if isinstance(other, (Decimal, float, int, str)):
            # The usual use-case, e.g. `5 * µ`
            return Prefixed(number=other, prefix=self)

        if isinstance(other, Prefixed):
            # Prefixed times Prefix, e.g. `(5 * n) * G`
            targ = self.value + other.prefix.value
            exptemp = e(targ)

            # Scale the other number
            new_num = other.number * Decimal(10) ** (targ - exptemp.symbol.value)

            # And create a corresponding `Prefixed`
            return Prefixed.new(new_num, exptemp.symbol)

        return NotImplemented

    def __truediv__(self, other: Any):
        """Division operator, e.g. `p / µ == µ`
        typical usage to evaluate prefix arithmetic"""

        if isinstance(other, Prefix):

            # Prefix divided by Prefix, eg. `p / M == a`
            targ = self.value - other.value
            return e(targ)

        return NotImplemented

    def __pow__(self, other: Any):
        """Power operator, e.g. `K ** 2 == M
        typical usage to evaluate prefix raised to integer powers"""

        if isinstance(other, (str, int, float, Decimal)):
            # Prefix raised to power of number, eg. `µ ** 4 == y`
            targ = self.value * Decimal(str(other))
            return e(targ)

        return NotImplemented

    def __int__(self):
        return int(10**self.value)

    def __float__(self):
        return float(10**self.value)

    def __str__(self):
        return str(Decimal(10) ** self.value)


"""
# Note on Numeric Types 

`Prefixed` supports a sole type for its `number` field: the standard library's `Decimal`. 
One might wonder why it does not include `int` and `float` as well, or perhaps 
some other numeric-like types. 

This boils down to limitations in validation. Previous versions of `Prefixed` have 
used a `Number` union-type along these lines:
```
Number = Union[Decimal, float, int]
```
This is nominally handy, but results in many values being converted among these types, 
often where one may not expect. 
The pydantic docs are clear about their limitations in this respect: 
https://pydantic-docs.helpmanual.io/usage/types/#unions

These limitations include the fact that despite being declared in list-like 
"ordered" syntax, union-types do not have orders in the Python standard library. 
So even interpreting `Union[Decimal, float, int]` as "prefer Decimal, use float 
and then int if that (for whatever reason) doesn't work" fails, 
since the `Union` syntax can be reordered arbitrarily by the language. 

The other clear alternative is not doing runtime type-validation of `Prefixed`, 
or of classes which instantiate them. Prior versions also tried this, 
to fairly confounding results. Incorporating standard-library `dataclasses` 
as members of larger `pydantic.dataclasses` seems to *work* - i.e. it does not 
produce `ValidationError`s or similar - but ultimately with enough of them, 
triggers some highly inscrutable errors in the standard-library methods. 

So: all `Decimal`, all `pydantic.dataclasses`.
"""


class Prefixed(BaseModel):
    """
    # Prefixed

    Combination of a literal value and a unit-indicating prefix.
    Colloquially, the numbers in expressions like "5ns", "11MV", and "1µA"
    are represented as `Prefixed`.
    """

    # Numeric Portion. See the long note above.
    number: Decimal
    # Enumerated SI Prefix. Defaults to unity.
    prefix: Prefix = Field(default=Prefix.UNIT)

    @classmethod
    def new(cls, number: Decimal, prefix: Prefix = Prefix.UNIT) -> "Prefixed":
        """Create a new Prefixed number.
        Alias for `Prefixed(number=number, prefix=prefix), and a (slight) shorthand
        for using arguments by-position, which `pydantic.BaseModel` does not support."""
        return cls(number=number, prefix=prefix)

    #
    # FIXME: pending potential deprecation #157
    #
    # @classmethod
    # def validate(cls, v: Union["Prefixed", "ToPrefixed"]) -> "Prefixed":
    #     """Validate `v` as a `Prefixed` number, or convert to `Prefixed` if applicable.
    #     While usable elsewhere, `validate` is primarily intended for use in type-validated
    #     dataclass trees, such as those generated in `paramclass`es."""
    #
    #     return to_prefixed(v)
    #
    # @classmethod
    # def __get_validators__(cls):
    #     yield cls.validate

    def __hash__(self):
        return hash((self.number, self.prefix))

    def __int__(self) -> int:
        return int(self.number) * 10**self.prefix.value

    def __float__(self) -> float:
        """Convert to float"""
        return float(self.number) * 10**self.prefix.value

    def __neg__(self) -> "Prefixed":
        return Prefixed.new(-self.number, self.prefix)

    def __abs__(self) -> "Prefixed":
        return Prefixed.new(abs(self.number), self.prefix)

    def __mul__(self, other) -> "Prefixed":
        if isinstance(other, Prefixed):
            return (self.number * other.number * self.prefix * other.prefix).scale()
        elif not isinstance(other, (str, int, float, Decimal)):
            return NotImplemented
        return Prefixed.new(self.number * Decimal(str(other)), self.prefix).scale()

    def __rmul__(self, other) -> "Prefixed":
        if isinstance(other, Prefixed):
            return (self.number * other.number * self.prefix * other.prefix).scale()
        elif not isinstance(other, (str, int, float, Decimal)):
            return NotImplemented
        return Prefixed.new(self.number * Decimal(str(other)), self.prefix).scale()

    def __truediv__(self, other) -> "Prefixed":
        if isinstance(other, Prefixed):
            return ((self.number / other.number) * (self.prefix / other.prefix)).scale()
        elif not isinstance(other, (str, int, float, Decimal)):
            return NotImplemented
        elif float(other) == 0.0:
            return float("inf")
        return Prefixed.new(
            self.number / Decimal(str(other)), self.prefix.value
        ).scale()

    def __rtruediv__(self, other) -> "Prefixed":
        if isinstance(other, Prefixed):
            return ((self.number / other.number) * (self.prefix / other.prefix)).scale()
        elif not isinstance(other, (str, int, float, Decimal)):
            return NotImplemented
        elif float(self.number) == 0.0:
            return float("inf")
        return Prefixed.new(
            Decimal(str(other)) / self.number, -self.prefix.value
        ).scale()

    def __pow__(self, other) -> "Prefixed":
        if not isinstance(other, (str, int, float, Decimal)):
            return NotImplemented
        return (
            self.number ** Decimal(str(other)) * (self.prefix ** Decimal(str(other)))
        ).scale()

    def __rpow__(self, other) -> "Prefixed":
        if not isinstance(other, (str, int, float, Decimal)):
            return NotImplemented
        return Prefixed.new(
            Decimal(str(other))
            ** (self.number * (10 ** Decimal(str(self.prefix.value))))
        )

    def __add__(self, other: "Prefixed") -> "Prefixed":
        if not isinstance(other, (str, int, float, Decimal, Prefixed)):
            return NotImplemented
        elif not isinstance(other, Prefixed):
            return _add(lhs=self, rhs=Prefixed.new(other))
        return _add(lhs=self, rhs=other).scale()

    def __radd__(self, other: "Prefixed") -> "Prefixed":
        if not isinstance(other, (str, int, float, Decimal, Prefixed)):
            return NotImplemented
        elif not isinstance(other, Prefixed):
            return _add(lhs=self, rhs=Prefixed.new(other))
        return _add(lhs=self, rhs=other).scale()

    def __sub__(self, other: "Prefixed") -> "Prefixed":
        if not isinstance(other, (str, int, float, Decimal, Prefixed)):
            return NotImplemented
        elif not isinstance(other, Prefixed):
            return _subtract(lhs=self, rhs=Prefixed.new(other))
        return _subtract(lhs=self, rhs=other).scale()

    def __rsub__(self, other: "Prefixed") -> "Prefixed":
        if not isinstance(other, (str, int, float, Decimal, Prefixed)):
            return NotImplemented
        elif not isinstance(other, Prefixed):
            return _subtract(lhs=Prefixed.new(other), rhs=self)
        return _subtract(lhs=other, rhs=self).scale()

    def scale(self, prefix: Prefix = None) -> "Prefixed":
        """Scale to a new `Prefix`"""
        if isinstance(prefix, Prefix):
            newnum = self.number * Decimal(10) ** (self.prefix.value - prefix.value)
            return Prefixed.new(newnum, prefix)
        else:
            newpref = Prefix.closest(abs(self.number).log10() + self.prefix.value)
            return self.scale(newpref)

    def __str__(self) -> str:
        return f"{self.number}*{self.prefix.name}"

    def __repr__(self) -> str:
        return f"{self.number}*{self.prefix.name}"

    # Comparison operators that respect class convention
    def __lt__(self, other) -> bool:
        lhs, rhs = _scale_to_smaller(self, other)
        return round(lhs.number, EPSILON) < round(rhs.number, EPSILON)

    def __le__(self, other) -> bool:
        lhs, rhs = _scale_to_smaller(self, other)
        return round(lhs.number, EPSILON) <= round(rhs.number, EPSILON)

    def __eq__(self, other) -> bool:
        lhs, rhs = _scale_to_smaller(self, other)
        return round(lhs.number, EPSILON) == round(rhs.number, EPSILON)

    def __ne__(self, other) -> bool:
        lhs, rhs = _scale_to_smaller(self, other)
        return round(lhs.number, EPSILON) != round(rhs.number, EPSILON)

    def __gt__(self, other) -> bool:
        lhs, rhs = _scale_to_smaller(self, other)
        return round(lhs.number, EPSILON) > round(rhs.number, EPSILON)

    def __ge__(self, other) -> bool:
        lhs, rhs = _scale_to_smaller(self, other)
        return round(lhs.number, EPSILON) >= round(rhs.number, EPSILON)


# Union of the types which can be converted to `Prefixed`
ToPrefixed = Union[int, float, str, Decimal]


def to_prefixed(v: Union[Prefixed, ToPrefixed]) -> Prefixed:
    """Convert any convertible type to a `Prefixed` number."""

    if isinstance(v, Prefixed):
        return v  # Valid as-is, return it.
    if isinstance(v, Decimal):
        return Prefixed(number=v)  # Also pretty much done
    # Convert the remaining convertible types to `Decimal` inline
    if isinstance(v, (int, float)):
        # Note that, like `pydantic`, we convert numeric types to `str` before passing to `Decimal`.
        return Prefixed(number=Decimal(str(v)))
    if isinstance(v, str):
        return Prefixed(number=Decimal(v))

    raise RuntimeError(f"Cannot convert {v} to Prefixed number")


def _add(lhs: Prefixed, rhs: Prefixed) -> Prefixed:
    """`Prefixed` Addition"""
    if lhs.prefix == rhs.prefix:
        return Prefixed.new(lhs.number + rhs.number, lhs.prefix)

    # Different prefix values. Scale to the smaller of the two
    smaller = lhs.prefix if lhs.prefix.value < rhs.prefix.value else rhs.prefix
    newnum = lhs.scale(smaller).number + rhs.scale(smaller).number
    return Prefixed.new(newnum, smaller)


def _subtract(lhs: Prefixed, rhs: Prefixed) -> Prefixed:
    """`Prefixed` Subtraction"""
    if lhs.prefix == rhs.prefix:
        return Prefixed.new(lhs.number - rhs.number, lhs.prefix)

    # Different prefix values. Scale to the smaller of the two
    smaller = lhs.prefix if lhs.prefix.value < rhs.prefix.value else rhs.prefix
    newnum = lhs.scale(smaller).number - rhs.scale(smaller).number
    return Prefixed.new(newnum, smaller)


def _scale_to_smaller(
    me: Prefixed, other: Union[Prefixed, ToPrefixed]
) -> Tuple[Prefixed, Prefixed]:
    """# Scale two `Prefixed` numbers to the smaller of the two prefixes.
    The `me` argument is always a `Prefixed`, generally due to being the `self` in a `Prefixed` mthod.
    The `other` argument is commonly another compatible/ convertible type,
    and is converted before scaling."""

    other = to_prefixed(other)
    smaller = (
        me.prefix
        if me.number * Decimal(10**me.prefix.value)
        < other.number * Decimal(10**other.prefix.value)
        else other.prefix
    )
    return me.scale(smaller), other.scale(smaller)


# Common prefixes as single-character identifiers, and exposed in the module namespace.
y = YOCTO = Prefix.YOCTO
z = ZEPTO = Prefix.ZEPTO
a = ATTO = Prefix.ATTO
f = FEMTO = Prefix.FEMTO
p = PICO = Prefix.PICO
n = NANO = Prefix.NANO
µ = u = MICRO = Prefix.MICRO
m = MILLI = Prefix.MILLI
c = CENTI = Prefix.CENTI
d = DECI = Prefix.DECI
D = DECA = Prefix.DECA
H = HECTO = Prefix.HECTO
K = KILO = Prefix.KILO
M = MEGA = Prefix.MEGA
G = GIGA = Prefix.GIGA
T = TERA = Prefix.TERA
P = PETA = Prefix.PETA
E = EXA = Prefix.EXA
Z = ZETTA = Prefix.ZETTA
Y = YOTTA = Prefix.YOTTA

# The Unit prefix doesn't get a single-character name, since it's kinda confusing with `µ`,
# but is exposed at module scope.
UNIT = Prefix.UNIT


@dataclass
class Exponent:
    """
    # Exponent

    Exponent is a helper class that aids Prefixed arithmetic in a way that is intuitive
    for users, such as allowing arbitrary powers, division and other functionality.

    A important feature of the design of Exponent is that it is a helper class: so it may
    only be invoked by other classes and can only interact with itself and Decimal/int/floats.
    """

    symbol: Prefix  # Prefix symbol for calculation
    residual: Decimal = Decimal(0)  # Prefix

    @classmethod
    def from_exp(self, exp: Any):

        out_symbol = Prefix.closest(exp)
        out_residual = Decimal(str(exp)) - out_symbol.value
        return Exponent(out_symbol, out_residual)

    def __mul__(self, other: Any) -> Optional[Union["Exponent", "Prefixed"]]:

        if isinstance(other, Exponent):
            # Exponent(n,0.3) * Exponent(K,0.4) == e(-8.7) * e(3.4) = e(-5.3)
            return Exponent.from_exp(
                self.symbol.value + other.symbol.value + self.residual + other.residual
            )

        elif isinstance(other, (str, int, float, Decimal)):
            return Prefixed.new(
                Decimal(str(other)) * Decimal(10) ** self.residual, self.symbol
            )

        return NotImplemented

    def __rmul__(self, other: Any) -> Prefixed:

        if isinstance(other, (str, int, float, Decimal)):
            # 16 * Exponent(Symbol.UNIT,0.25) == 2 * Prefix.UNIT

            out_number = Decimal(str(other)) * Decimal(10) ** self.residual

            return Prefixed.new(out_number, self.symbol)

        elif isinstance(other, Prefixed):
            # 2 * Prefix.UNIT * Exponent(Symbol.KILO,0) == 2 * Prefix.KILO
            temp_exp = e(other.prefix.value)
            temp_exp *= self
            return other.number * temp_exp

        return NotImplemented

    def __truediv__(self, other: Any) -> Optional["Exponent"]:

        if isinstance(other, (str, int, float, Decimal)):
            # Exponent(1) / 2 = Exponent(log(5))
            inv_other = Decimal(1) / Decimal(str(other))
            return inv_other * self

        elif isinstance(other, Exponent):
            return self * (other**-1)

        return NotImplemented

    def __pow__(self, other: Any) -> Optional["Exponent"]:

        if isinstance(other, (str, int, float, Decimal)):
            # Exponent(0.25) ** 4 == Exponent(1)
            return Exponent.from_exp(
                ((self.symbol.value + self.residual) * Decimal(str(other)))
            )

        return NotImplemented

    def __repr__(self) -> str:
        return f"e({self.symbol.value+self.residual})"

    def __float__(self) -> float:
        return float(10 ** (self.symbol.value + self.residual))

    def __str__(self) -> str:
        return str(Decimal(10) ** (self.symbol.value + self.residual))


def e(exp: Any) -> Exponent:
    """# Exponential `Prefix` Creation

    Returns an `Exponent` for power-of-ten exponent `exp`.

    In many cases HDL parameters must be non-integer values,
    e.g. `1nm`, but using `float` can prove undesirable
    as rounding errors can turn them into subtly different values.
    The `e()` function is most commonly useful with multiplication,
    to create "floating point" values such as `11 * e(-9)`.
    """

    if isinstance(exp, (str, int, float, Decimal)):

        out_symbol = Prefix.closest(exp)
        out_residual = Decimal(str(exp)) - out_symbol.value

        return Exponent(out_symbol, out_residual)

    return NotImplemented


# Star-imports *do not* include the single-character names `µ`, `e`, et al.
# They can be explicityle imported from `hdl21.prefix` instead.
__all__ = ["Prefix", "Prefixed", "Exponent"]



"""
# Instance
Hierarchical instance of another `Module`, `Primitive`, or similar piece of replicable hardware.
"""

# Std-Lib Imports
from typing import Optional, Any, Dict, Type, TypeVar
from dataclasses import dataclass, field
from textwrap import dedent

# Local imports
from .source_info import source_info, SourceInfo
from .attrmagic import init
from .connect import Connectable, is_connectable
from .props import Properties

T = TypeVar("T")


@init
class _Instance:
    """# _Instance
    Shared base class for Instance-like types (Instance, InstanceArray)"""

    def __init__(
        self,
        of: "Instantiable",
        *,
        name: Optional[str] = None,
    ):
        from .instantiable import Instantiable, is_instantiable

        if not is_instantiable(of):
            raise RuntimeError(f"Invalid Instance of {of}")

        self.name: Optional[str] = name
        self.of: Instantiable = of
        self.conns: Dict[str, "Connectable"] = dict()
        self.props: Properties = Properties()

        # References we give out, either for refering to ports or entries in out own `conns`
        self._refs = Refs()

        self._parent_module: Optional["Module"] = None  # Instantiating module
        self._elaborated: bool = False
        self._source_info: Optional[SourceInfo] = source_info(get_pymodule=False)

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name} of={self.of})"

    def __getattr__(self, key: str) -> Any:
        """Port access by getattr"""
        if not self.__getattribute__("_initialized") or key.startswith("_"):
            # Bootstrapping phase: do regular getattrs to get started
            return object.__getattribute__(self, key)

        if key in self.__getattribute__("_specialcases"):  # Special case(s)
            return object.__getattribute__(self, key)

        # After elaboration, the fancy PortRef creation below goes away. Only return ready-made connections.
        if self.__getattribute__("_elaborated"):
            # Pull from the `conns` dict.
            conns = self.__getattribute__("conns")
            if key in conns.keys():
                return conns[key]
            raise AttributeError(f"No attribute {key} for {self}")

        # Fell through all those cases. Fancy `PortRef` generation time!
        # Return a `PortRef`, creating one if necessary.
        return _get_portref(self, key)

    def __call__(self, **kwargs) -> "_Instance":
        """Connect-by-call"""
        for key, val in kwargs.items():
            self.connect(key, val)
        # Don't forget to retain ourselves at the call-site!
        return self

    def __setattr__(self, key: str, val: Any) -> None:
        """Connect-by-setattr"""
        if not getattr(self, "_initialized", False) or key.startswith("_"):
            # Bootstrapping phase: do regular setattrs to get started
            return object.__setattr__(self, key, val)
        if key in self.__getattribute__("_specialcases"):  # Special case(s)
            return object.__setattr__(self, key, val)
        _ = self.connect(key, val)  # Discard the returned `self`
        return None

    def connect(self, portname: str, conn: Connectable) -> "_Instance":
        """Connect `conn` to port (name) `portname`.
        Called by both by-call and by-assignment convenience methods, and usable directly.
        Direct calls to `connect` will generally be required for ports with otherwise illegal names,
        e.g. Python language keywords (`in`, `from`, etc.),
        or Hdl21 internal "keywords" (`name`, `ports`, `signals`, etc.).
        Returns `self` to aid in method-chaining use-cases."""
        from .bundle import AnonymousBundle

        if isinstance(conn, Dict):
            # Special-case dictionaries of connectables into Anon Bundles
            conn = AnonymousBundle(**conn)
        if not is_connectable(conn):
            raise TypeError(f"{self} attempting to connect non-connectable {conn}")

        # The main event: actually stick `conn` in the `conns` dict
        if portname in self.conns:
            # Replace and disconnect any prior connection. Disregards the returned old connection.
            self.replace(portname, conn)
        else:
            self.conns[portname] = conn
            conn._connected_ports.add(_get_connref(self, portname))

        # And return `self` to aid in method-chaining use-cases
        return self

    def disconnect(self, portname: str) -> Connectable:
        """Disconnect the port named `portname`.
        Returns the formerly-connected `Connectable`.
        Raises a KeyError if the port is not connected."""

        conn = self.conns.pop(portname)
        conn._connected_ports.remove(_get_connref(self, portname))
        return conn

    def replace(self, portname: str, conn: Connectable) -> Connectable:
        """
        Replace the connection to `portname` with `conn`.
        Returns the formerly-connected `Connectable`.
        Raises a KeyError if the port is not connected.

        The `replace` method is functionally identical to serial calls to `disconnect` and `connect`,
        but allows for in-place modification of the `conns` dict, e.g. while iterating over its items.
        """

        connref = _get_connref(self, portname)
        # Get a reference to the old connection in the `conns` dict, without removing it
        old = self.conns[portname]
        old._connected_ports.remove(connref)
        # And replace it in the `conns` dict
        self.conns[portname] = conn
        conn._connected_ports.add(connref)
        return old


def _mult(inst: "Instance", other: int) -> "InstanceArray":
    """Instance by integer multiplication.
    Creates an Instance Array of size `other`."""
    if not isinstance(other, int):
        return NotImplemented
    return _to_array(inst=inst, num=other)


class Instance(_Instance):
    """
    # Instance
    Hierarchical instance of another `Module`, `Primitive`, or similar piece of replicable hardware.
    """

    __mul__ = __rmul__ = _mult  # Apply `_mult` on both left and right

    _specialcases = [
        "name",
        "of",
        "conns",
        "connect",
        "disconnect",
        "replace",
    ]

    def __init__(self, *_, **__):
        # Just call the base-class constructor, and set the `_initialized` flag.
        # The need to do so is just an enabler for other sub-classes to add constructor action.
        super().__init__(*_, **__)
        self._initialized = True


class InstanceBundle(_Instance):
    """# Instance Bundle
    A named set of Instances, paired with a Signal Bundle."""

    # The paired signal-Bundle type. Note this is a class-level attribute.
    bundle: Optional["Bundle"] = None

    _specialcases = [
        "name",
        "of",
        "conns",
        "connect",
        "disconnect",
        "replace",
    ]

    def __init__(self, *_, **__):
        # Just call the base-class constructor, and set the `_initialized` flag.
        # The need to do so is just an enabler for other sub-classes to add constructor action.
        super().__init__(*_, **__)
        self._initialized = True


def InstanceBundleType(name: str, bundle: "Bundle", doc: Optional[str] = None) -> type:
    """Create a new sub-class of `InstanceBundle`, tied to Bundle-type `bundle`."""

    from .bundle import Bundle

    if not isinstance(name, str):
        raise TypeError
    if not isinstance(bundle, Bundle):
        raise TypeError

    # Set a default doc-string
    doc = doc or f"`InstanceBundle` type `{name}` of `Bundle` `{bundle.name}`"
    doc = dedent(doc)

    # Create and return the new type
    return type(name, (InstanceBundle,), {"bundle": bundle, "__doc__": doc})


class InstanceArray(_Instance):
    """# Instance Array
    A single-dimensional array of `n` `Instance`s of the same `Module`.
    Enables numpy-style "connection broadcasting" semantics."""

    _specialcases = [
        "name",
        "of",
        "conns",
        "portref",
        "portrefs",
        "connect",
        "disconnect",
        "replace",
    ]

    def __init__(
        self,
        of: "Instantiable",
        n: int,
        name: Optional[str] = None,
    ):
        super().__init__(of=of, name=name)
        self.n = n
        self._initialized = True

    def __getitem__(self, idx: int):
        return RuntimeError(f"Illegal indexing into Array {self}")

    def __setitem__(self, _idx: Any, _val: Any):
        return RuntimeError(f"Illegal indexing into Array {self}")


""" 
Module-level functions 
These could be instance methods, but are kept here to minimally pollute their namespaces.
"""


def _to_array(inst: Instance, num: int) -> InstanceArray:
    """Create an Instance Array from an Instance"""
    # Several constraints asserted here which may eventually be relaxed.
    # * No port-references (yet)
    # * Not a member of a module (yet)
    if len(inst._refs.portrefs) > 0:
        msg = f"Cannot convert Instance {inst} with outstanding port-references {inst._portrefs} to Array"
        raise RuntimeError(msg)
    if inst._parent_module is not None:
        msg = f"Cannot convert Instance {inst} already inserted in Module {inst._parent_module} to Array"
        raise RuntimeError(msg)

    # Checks out. Create the array.
    return InstanceArray(of=inst.of, n=num, name=inst.name)(**inst.conns)


@dataclass
class Refs:
    """Tracking of references stored on each Instance.
    All references are of type `PortRef`, and are organized into two camps:

    * The `portrefs` dict includes entries like `i.someport` below:
    ```
    i = Instance(of=MyModule)
    i2 = Instance(of=AnotherModule)(itsport=i.someport)
    ```
    These are ultimately resolved to Signals during elaboration.

    * The `conns` dict includes entries like `i.someport` below:
    ```
    s = Signal()
    i = Instance(of=MyModule)(someport=s)
    ```
    Here the Signal `s` is given a port-reference to `i`, to help track and update later connections.
    These entries are members of the `connrefs` dict.

    The combination of the two is stored in the `all` dictionary.
    All items in both `portrefs` and `connrefs` are always also located in `all`.
    Attempts to retrieve a reference come from `all`, and therefore generate the same
    objects for successive fetches of the same port name.

    Entries are never removed from `refs` over the course of an Instance's lifetime.
    For most Instances, by the end of their life, the `connrefs` and `all` dicts will contain
    an entry for each of its ports.
    """

    all: Dict[str, "PortRef"] = field(default_factory=dict)
    portrefs: Dict[str, "PortRef"] = field(default_factory=dict)
    connrefs: Dict[str, "PortRef"] = field(default_factory=dict)


def _get_portref(self: _Instance, key: str) -> "PortRef":
    """Return a port-reference to name `key`, creating it if necessary."""
    from .portref import PortRef

    # Check in our existing references
    refs = self.__getattribute__("_refs")
    if key in refs.all:
        refs.portrefs[key] = refs.all[key]
        return refs.all[key]

    # New reference; create, add, and return it
    ref = PortRef(inst=self, portname=key)
    refs.portrefs[key] = refs.all[key] = ref
    return ref


def _get_connref(self: _Instance, key: str) -> "PortRef":
    """Return a connection-reference to name `key`, creating it if necessary."""
    from .portref import PortRef

    # Check in our existing references
    refs = self.__getattribute__("_refs")
    if key in refs.all:
        refs.connrefs[key] = refs.all[key]
        return refs.all[key]

    # New reference; create, add, and return it
    ref = PortRef(inst=self, portname=key)
    refs.connrefs[key] = refs.all[key] = ref
    return ref


""" 
Instantiation Decorator 
"""


def calls_instantiate(cls: Type[T]) -> Type[T]:
    """# Calls Instantiate
    Decorator which adds 'calls produce `hdl21.Instance`s' functionality.
    Added to everything that is `Instantiable`, notably including `Module`."""

    def __call__(self, **kwargs) -> Instance:
        """Calls Create `hdl21.Instances`,
        and pass any (keyword-only) arguments to said `Instances`,
        generally to connect-by-call."""
        return Instance(of=self)(**kwargs)

    # Check for an existing __call__ method, and if there is one, bail
    if "__call__" in cls.__dict__:
        msg = f"Hdl21 Internal Error: Invalid conflict between `calls_instantiate` decorator and explicit `__call__` method on {cls}"
        raise RuntimeError(msg)

    cls.__call__ = __call__
    return cls


"""
Selected star-exports 
Notably excludes the base class, and module-level functions
"""

__all__ = [
    "Instance",
    "InstanceArray",
    "InstanceBundle",
    "InstanceBundleType",
    "calls_instantiate",
]



"""
# Qualified Names 

Helpers for uniquely identifying a Module or ExternalModule, 
given their native language (Python) has hierarchical namespaces, 
and many of their import/ export languages (e.g. Verilog, netlists) do not.
"""

from typing import Optional, Union, List

# Type alias of the types that have a qualified path.
HasQualPath = Union["Module", "ExternalModule", "Generator"]


def qualpath(mod: HasQualPath) -> Optional[List[str]]:
    """# Qualified Path
    Helper for exporting. Returns a module's definition path.
    This is generally one of a few things:
    * If "normally" defined via Python code, it's the Python module path plus the module name.
    * If *imported*, it's the path inferred during import."""

    if getattr(mod, "_importpath", None) is not None:
        # Imported. Return the period-separated import path.
        return mod._importpath + [mod.name]

    if mod.name is None:
        # Unnamed. Return None.
        return None

    if mod._source_info.pymodule is None:
        # Defined outside a Python module, e.g. in a call to `exec`, a notebook cell, or a `python -c` string.
        # Return its name without any path qualifiers.
        return [mod.name]

    # Defined the old fashioned way. Use the Python module name.
    return mod._source_info.pymodule.__name__.split(".") + [mod.name]


def qualname(mod: HasQualPath) -> Optional[str]:
    """# Qualified Name
    Helper for exporting. Returns a module's path-qualified name.
    If `mod` has a qualified path as determined by `qualpath`, returns it
    joined together by the Python-conventional path-separator "."."""

    qpath = qualpath(mod)
    if qpath is None:
        return None
    return ".".join(qpath)



"""
# Hierachy Walker Module 
"""

# Std-Lib
from typing import List

# Local imports
from .elab import Elaboratable, Elaboratables, elaborate, is_elaboratable
from .instance import Instance
from .instantiable import Instantiable
from .module import Module
from .external_module import ExternalModuleCall
from .primitives import PrimitiveCall


class HierarchyWalker:
    """
    # Hierarchical Walker

    Walks a hierarchical design tree.
    Designed to be used as a base-class for extensions such as process-specific instance-replacements.
    """

    def visit_elaboratables(self, src: Elaboratables) -> Elaboratables:
        """Visit an `Elaboratables` object.
        Largely dispatches across the type-union of elaboratable objects."""

        # First ensure all the `src` modules and generators are elaborated.
        # This is a functional no-op if they already are.
        elaborate(src)

        if isinstance(src, List):
            for x in src:
                self.visit_elaboratable(x)
            return src
        if is_elaboratable(src):
            return self.visit_elaboratable(src)
        raise TypeError(f"Cannot walk non-elaboratable {src}")

    def visit_elaboratable(self, src: Elaboratable) -> Elaboratable:
        """Visit an `Elaboratable` object."""

        if isinstance(src, Module):
            return self.visit_module(src)
        if isinstance(src, ExternalModuleCall):
            return self.visit_external_module_call(src)
        if isinstance(src, PrimitiveCall):
            return self.visit_primitive_call(src)
        raise TypeError(f"Cannot walk non-elaboratable {src}")

    def visit_module(self, module: Module) -> Instantiable:
        """Visit a `Module`.
        Primary method for most manipulations."""

        # Step into each of the Module's instances.
        # Note that as we have already elaborated, it no longer has bundles.
        for inst in module.instances.values():
            self.visit_instance(inst)
        return module

    def visit_external_module_call(self, call: ExternalModuleCall) -> Instantiable:
        """Visit an `ExternalModuleCall`.
        A common place to perform "module replacement" in sub-classes.
        Base implementation does nothing."""
        return call

    def visit_primitive_call(self, call: PrimitiveCall) -> Instantiable:
        """Visit an `PrimitiveCall`.
        A common place to perform "module replacement" in sub-classes,
        especially when transforming Hdl21's generic `Primitive`s into technology-specific devices.
        Base implementation does nothing."""
        return call

    def visit_instance(self, inst: Instance) -> Instance:
        """Visit a hierarchical `Instance`.
        Visits and sets its `of` field to the result of `visit_instantiable`."""
        inst.of = self.visit_instantiable(inst.of)
        return inst

    def visit_instantiable(self, of: Instantiable) -> Instantiable:
        """Visit an `Instantiable` instance-target."""

        if isinstance(of, Module):
            return self.visit_module(of)
        if isinstance(of, PrimitiveCall):
            return self.visit_primitive_call(of)
        if isinstance(of, ExternalModuleCall):
            return self.visit_external_module_call(of)
        raise TypeError(f"Invalid Instance of {of}")

    @classmethod
    def walk(cls, src: Elaboratables) -> Elaboratables:
        """Walk a hierarchical design tree or list of them.
        Class-level method commonly called on sub-classes, e.g. in PDK compilation methods."""
        return cls().visit_elaboratables(src)


def walk(src: Elaboratables) -> Elaboratables:
    """Walk a hierarchical design tree or list of them."""
    return HierarchyWalker.walk(src)



"""
# Array Flattening 
"""

# Local imports
from ...module import Module
from ...instance import Instance
from ...portref import PortRef
from ...bundle import BundleInstance
from ...signal import Signal
from ...slice import Slice
from ...concat import Concat

# Import the base class
from .base import ElabPass


class ArrayFlattener(ElabPass):
    """
    Elaboration Pass to Flatten `InstanceArray`s into `Instance`s, broadcast and remake their connections.
    """

    def elaborate_module(self, module: Module) -> Module:
        """Elaborate Module `module`.
        Primarily performs flattening of Instance Arrays, and re-connecting to the resultant flattened instances."""

        # Flatten Instance arrays
        while module.instarrays:
            name, array = module.instarrays.popitem()
            self.stack.append(array)
            module.namespace.pop(name)

            # Visit the array's target
            target = self.elaborate_instance_base(array)

            # And do the real work: flattening it.
            if array.n < 1:
                self.fail(f"Invalid InstanceArray {array} with size {array.n}")

            # Create the new, flat Instances
            new_insts = []
            for k in range(array.n):
                name = self.flatname(
                    segments=[array.name, str(k)], avoid=module.namespace
                )
                inst = module.add(Instance(of=target, name=name))
                new_insts.append(inst)

            # And connect them
            for portname, conn in array.conns.items():
                if isinstance(conn, BundleInstance):
                    # All new instances get the same BundleInstance
                    for inst in new_insts:
                        inst.connect(portname, conn)
                elif isinstance(conn, (Signal, Slice, Concat)):
                    # Get the target-module port, particularly for its width
                    port = target.ports.get(portname, None)
                    if port is None:
                        msg = f"Connection to invalid Port `{portname}` on InstanceArray `{array}` in Module `{module.name}`"
                        self.fail(msg)
                    if not isinstance(port, Signal):
                        msg = f"Invalid Port `{portname}` ({port}) on InstanceArray `{array}` in Module `{module.name}`"
                        self.fail(msg)

                    if port.width == conn.width:
                        # All new instances get the same signal
                        for inst in new_insts:
                            inst.connect(portname, conn)
                    elif port.width * array.n == conn.width:
                        # Each new instance gets a slice, equal to its own width
                        for k, inst in enumerate(new_insts):
                            slize = conn[k * port.width : (k + 1) * port.width]
                            if slize.width != port.width:
                                msg = f"Width mismatch connecting {slize} to {port}"
                                self.fail(msg)
                            inst.connect(portname, slize)
                    else:  # All other width-values are invalid
                        msg = f"Invalid connection of {conn} of width {conn.width} to port {portname} on Array {array.name} of width {port.width}. "
                        msg += f"Valid widths are either {port.width} (broadcasting across instances) and {port.width * array.n} (individually wiring to each)."
                        self.fail(msg)
                elif isinstance(conn, PortRef):
                    msg = f"Error elaborating {array} in {module}. "
                    msg += f"Connection {conn} has not been resolved to a `Signal`. "
                    raise RuntimeError
                else:
                    msg = f"Invalid connection to {conn} in InstanceArray {array}"
                    self.fail(msg)
            self.stack.pop()

        return module



"""
# `hdl21.sim` Unit Tests
"""

import pytest

import hdl21 as h
from hdl21.sim import *
from hdl21.prefix import m
from hdl21.primitives import Vdc
import vlsir.spice_pb2 as vsp
import vlsirtools
from vlsirtools.spice import sim, SimOptions, ResultFormat, sim_data as sd


def test_sim1():
    """Test minimal `Sim` creation"""
    s = Sim(tb=tb("empty"), attrs=[])
    assert isinstance(s, Sim)
    assert s.tb.name == "empty"
    assert len(s.tb.ports) == 1
    assert s.attrs == []


@h.module
class MyTb:
    # Create a sample testbench, particularly with enough stuff in it for `Noise` analysis.
    # I.e. a source instance and a non-ground Signal.
    VSS = h.Port()
    p = h.Signal()
    v = Vdc(dc=0 * m, ac=1000 * m)(p=p, n=VSS)


def test_sim2():
    """Test creating a more fully-featured sim"""
    s = Sim(
        tb=MyTb,
        attrs=[
            Param(name="x", val=5),
            Dc(var="x", sweep=PointSweep([1]), name="mydc"),
            Ac(sweep=LogSweep(1e1, 1e10, 10), name="myac"),
            Tran(tstop=11 * h.prefix.p, name="mytran"),
            Noise(
                output=MyTb.p,
                input_source=MyTb.v,
                sweep=LogSweep(1e1, 1e10, 10),
                name="mynoise",
            ),
            SweepAnalysis(
                inner=[Tran(tstop=1, name="swptran")],
                var="x",
                sweep=LinearSweep(0, 1, 2),
                name="mysweep",
            ),
            MonteCarlo(
                inner=[
                    Dc(var="y", sweep=PointSweep([1]), name="swpdc"),
                ],
                npts=11,
                name="mymc",
            ),
            Save(SaveMode.ALL),
            Meas(analysis="mytr", name="a_delay", expr="trig_targ_something"),
            Include("/home/models"),
            Lib(path="/home/models", section="fast"),
            Options(1e-9, name="reltol"),
        ],
    )
    to_proto(s)


def test_simattrs():
    """Test the "sim attrs" feature, which adds methods to `Sim` for each `SimAttr`"""

    s = Sim(tb=MyTb)

    p = s.param(name="x", val=5)
    dc = s.dc(var=p, sweep=PointSweep([1]), name="mydc")
    ac = s.ac(sweep=LogSweep(1e1, 1e10, 10), name="myac")
    tr = s.tran(tstop=11 * h.prefix.p, name="mytran")
    noise = s.noise(
        output=MyTb.p,
        input_source=MyTb.v,
        sweep=LogSweep(1e1, 1e10, 10),
        name="mynoise",
    )
    assert tr.tstop == 11 * h.prefix.p
    sw = s.sweepanalysis(inner=[tr], var=p, sweep=LinearSweep(0, 1, 2), name="mysweep")
    mc = s.montecarlo(
        inner=[
            Dc(var="y", sweep=PointSweep([1]), name="swpdc"),
        ],
        npts=11,
        name="mymc",
    )
    s.save(SaveMode.ALL)
    s.meas(analysis=tr, name="a_delay", expr="trig_targ_something")
    s.include("/home/models")
    s.lib(path="/home/models", section="fast")
    s.options(1e-9, name="reltol")

    to_proto(s)


def test_sim_decorator():
    """Test creating the same Sim, via the class decorator"""

    @h.sim.sim
    class MySim:
        tb = MyTb

        x = Param(5)
        y = Param(6)
        mydc = Dc(var=x, sweep=PointSweep([1]))
        myac = Ac(sweep=LogSweep(1e1, 1e10, 10))
        mytran = Tran(tstop=11 * h.prefix.p)
        mynoise = Noise(
            output=MyTb.p,
            input_source=MyTb.v,
            sweep=LogSweep(1e1, 1e10, 10),
        )
        mysweep = SweepAnalysis(
            inner=[mytran],
            var=x,
            sweep=LinearSweep(0, 1, 2),
        )
        mymc = MonteCarlo(
            inner=[
                Dc(var="y", sweep=PointSweep([1]), name="swpdc"),
            ],
            npts=11,
        )
        a_delay = Meas(analysis=mytran, expr="trig_targ_something")
        opts = Options(1e-9, name="reltol")

        # Attributes whose names don't really matter can be called anything,
        # but must be *assigned* into the class, not just constructed.
        _ = Save(SaveMode.ALL)

        # The `a_path` field will be dropped from the `Sim` definition,
        # but can be referred to by the following attributes.
        a_path = "/home/models"
        _ = Include(a_path)
        _ = Lib(path=a_path, section="fast")

    assert isinstance(MySim, Sim)
    assert MySim.name == "MySim"
    assert isinstance(MySim.tb, h.Module)
    assert MySim.tb.name == "MyTb"
    assert isinstance(MySim.attrs, list)
    for attr in MySim.attrs:
        assert is_simattr(attr)
    assert not hasattr(MySim, "a_path")

    to_proto(MySim)


def test_tb():
    """Test the `tb` function"""
    mytb = tb("mytb")
    assert isinstance(mytb, h.Module)
    assert is_tb(mytb)
    assert len(mytb.ports) == 1


def test_proto1():
    """Test exporting `Sim` to the VLSIR Protobuf schema"""

    s = Sim(
        tb=MyTb,
        attrs=[
            Param(name="x", val=5),
            Dc(var="x", sweep=PointSweep([1]), name="mydc"),
            Ac(sweep=LogSweep(1e1, 1e10, 10), name="myac"),
            Tran(tstop=11 * h.prefix.p, name="mytran"),
            SweepAnalysis(
                inner=[Tran(tstop=1, name="swptran")],
                var="x",
                sweep=LinearSweep(0, 1, 2),
                name="mysweep",
            ),
            MonteCarlo(
                inner=[
                    Dc(var="y", sweep=PointSweep([1]), name="swpdc"),
                ],
                npts=11,
                name="mymc",
            ),
            Save(SaveMode.ALL),
            Meas(analysis="mytr", name="a_delay", expr="trig_targ_something"),
            Include("/home/models"),
            Lib(path="/home/models", section="fast"),
            Options(1e-9, name="reltol"),
        ],
    )

    p = to_proto(s)

    import vlsir.circuit_pb2 as vckt
    import vlsir.spice_pb2 as vsp

    assert isinstance(p.pkg, vckt.Package)
    assert p.top == "test_sim.MyTb"


def test_generator_sim():
    """Test creating and exporting `Sim` with generator-valued DUTs,
    particularly several with different parameter-values."""

    @h.paramclass
    class P:  # A largely dummy param-class
        i = h.Param(dtype=int, desc="An integer", default=11)

    @h.generator
    def G(p: P) -> h.Module:
        m = h.Module()
        m.VSS = h.Port()
        return m

    # Create a few instances
    g1 = G(P(1))
    g2 = G(P(2))

    # Create `Sim`s of them
    s1 = Sim(tb=g1)
    s2 = Sim(tb=g2)

    # And export both to protobuf
    p1 = to_proto(s1)
    p2 = to_proto(s2)

    assert p1.top != p2.top
    assert p1.top == "test_sim.G(i=1)"
    assert p2.top == "test_sim.G(i=2)"


def test_delay1():
    from hdl21.sim import delay
    from hdl21.sim.delay import DelaySimParams, LogicState, Transition
    from hdl21.prefix import p, n, f, m

    @h.module
    class M:
        i0, i1 = h.Inputs(2)
        o0, o1 = h.Outputs(2)
        vdd, vss = h.Ports(2)

    p = DelaySimParams(
        dut=M,
        primary_input=M.i0,
        other_inputs=dict(i1=LogicState.HIGH),
        input_trans=Transition.RISING,
        vlo=0 * m,
        vhi=1000 * m,
        supplies=dict(vdd=1000 * m),
        grounds=[M.vss],
        load_caps=dict(o0=2 * f),
        default_load_cap=1 * f,
        tstop=1 * n,
        tstep=1 * p,
        pathsep=":",
    )
    delay_sim = delay.create_sim(p)
    h.sim.to_proto(delay_sim)
    # FIXME! some real checks plz


def empty_tb(num=0) -> h.Module:
    from hdl21.prefix import K
    from hdl21.primitives import R

    ri = R(R.Params(r=1 * K))

    @h.module
    class EmptyTb:
        """An Empty TestBench,
        or as close as we can get to one without some simulators failing.
        AKA, a resistor to ground."""

        VSS = h.Port()
        s = h.Signal()
        r = ri(p=s, n=VSS)

    EmptyTb.name = f"EmptyTb{num}"
    return EmptyTb


@pytest.mark.skipif(
    vlsirtools.spice.default() is None,
    reason="No simulator available",
)
def test_empty_sim1():
    """Create and run an empty `Sim`, returning a VLSIR_PROTO"""

    s = Sim(tb=empty_tb(), attrs=[])
    r = sim(to_proto(s), SimOptions(fmt=ResultFormat.VLSIR_PROTO))
    assert isinstance(r, vsp.SimResult)
    assert not len(r.an)  # No analysis inputs, no analysis results


@pytest.mark.skipif(
    vlsirtools.spice.default() is None,
    reason="No simulator available",
)
@pytest.mark.skipif(
    vlsirtools.spice.default() == vlsirtools.spice.SupportedSimulators.XYCE,
    reason="No support for `Xyce` + `SimData` python types",
)
def test_empty_sim2():
    """Create and run an empty `Sim`, returning SIM_DATA"""

    s = Sim(tb=empty_tb(), attrs=[])
    r = sim(to_proto(s), SimOptions(fmt=ResultFormat.SIM_DATA))
    assert isinstance(r, sd.SimResult)
    assert not len(r.an)  # No analysis inputs, no analysis results


@pytest.mark.xfail(reason="VLSIR #71 https://github.com/Vlsir/Vlsir/issues/71")
def test_multi_sim():
    """Test multiple Sims in parallel"""
    s1 = Sim(tb=empty_tb(1), attrs=[])
    s2 = Sim(tb=empty_tb(2), attrs=[])
    s3 = Sim(tb=empty_tb(3), attrs=[])
    s4 = Sim(tb=empty_tb(4), attrs=[])

    r = sim(to_proto([s1, s2, s3, s4]), SimOptions(fmt=ResultFormat.VLSIR_PROTO))

    for a in r:
        assert isinstance(a, vsp.SimResult)
        assert not len(a.an)


def really_empty_tb() -> h.Module:
    @h.module
    class ReallyEmptyTb:
        """An Empty TestBench, this time REALLY empty.
        For inducing an error in a multi-sim below."""

    return ReallyEmptyTb


def test_multi_sim_error():
    """Test that a failure in multiple concurrent sims fails, unlike in erdewit/nest_asyncio#57"""
    with pytest.raises(Exception):
        s1 = Sim(tb=empty_tb(1), attrs=[])
        s2 = Sim(tb=really_empty_tb(), attrs=[])
        s3 = Sim(tb=empty_tb(3), attrs=[])
        s4 = Sim(tb=empty_tb(4), attrs=[])

        r = sim(to_proto([s1, s2, s3, s4]), SimOptions(fmt=ResultFormat.VLSIR_PROTO))



"""
# Default Value Sentinel 

Used by a number of Hdl21 types to indicate a default value, 
where all other "public" values could be valid non-default data. 
"""

from .datatype import datatype


@datatype
class Default:
    """
    # Default Value Sentinel

    Used by a number of Hdl21 types to indicate a default value,
    where all other "public" values could be valid non-default data.
    """

    def __new__(cls, *_, **__):
        # And we're a singleton. Call away, you get the same class-object back.
        return Default



"""
Type-alias for `Instantiable` hdl21 types. Each is valid as the `of` field of `hdl21.Instance`s, 
and thus supports its "connect by call" and "connect by assignment" semantics. 
"""

# Std-Lib Imports
from __future__ import annotations
from typing import Any, Union, Dict
import copy

# Local Imports
from .datatype import AllowArbConfig, _pydantic_major_version
from .module import Module
from .primitives import PrimitiveCall
from .external_module import ExternalModuleCall


# Instantiable types-union
InstantiableUnion = Union[Module, ExternalModuleCall, PrimitiveCall]


def assert_instantiable(i: Any) -> "Instantiable":
    """# Assert that `i` is an `Instantiable` type."""
    if not is_instantiable(i):
        return invalid(i)
    return i


def is_instantiable(val: Any) -> bool:
    """Boolean indication of whether `val` is an `Instantiable` type."""
    return isinstance(val, InstantiableUnion.__args__)


def invalid(val: Any) -> None:
    """Raise a `TypeError` with debug info for invalid `Instantiable` `val`."""

    # Give more specific error-messages for our common types.
    # Especially those that are easily confused, e.g. `Primitive`, `ExternalModule`, and `Generator`
    from .external_module import ExternalModule
    from .primitives import Primitive
    from .generator import Generator
    from .instance import _Instance

    msg = f"Invalid `Instantiable` {val} of type {type(val)}.\n"
    if isinstance(val, (Generator, Primitive, ExternalModule)):
        msg += f"Did you mean to *call it* to apply parameters first?"
    elif isinstance(val, _Instance):
        msg += f"Did you mean to use its target module `{val.of}` instead?"
    else:  # Generic message for everything else
        msg += f"Valid `Instantiable` types include: {list(InstantiableUnion.__args__)}"
    raise TypeError(msg)


def qualname(i: "Instantiable") -> str:
    """Path-qualified name of Instantiable `i`"""
    from .qualname import qualname as module_qualname

    if isinstance(i, PrimitiveCall):
        # These have no "qualification" paths, just a singular name.
        return i.name

    # The other variants can have a path-qualifier
    if isinstance(i, Module):
        return module_qualname(i)
    if isinstance(i, ExternalModuleCall):
        return module_qualname(i.module)

    raise TypeError(f"Invalid Instantiable {i}")


def io(i: "Instantiable") -> Dict[str, "Connectable"]:
    """
    Get a complete dictionary of IO ports for `i`, including all types: Signals and Bundles.
    Copies the Instantiable's top-level dictionary so that it is not modified by consumers.
    """

    rv = copy.copy(i.ports)
    if hasattr(i, "bundle_ports"):
        rv.update(copy.copy(i.bundle_ports))
    return rv


_doc = """
# Instantiable

Generally this means
````python
Union[Module, ExternalModuleCall, PrimitiveCall]
```
with some customized checking and error handling.
"""

if _pydantic_major_version == 1:
    from .datatype import BaseModel

    class Instantiable(BaseModel):
        # "Custom root types" implementation

        __doc__ = _doc
        __root__: InstantiableUnion
        Config = AllowArbConfig

        def __init__(self, *_, **__):
            # Brick any attempts to create instances
            msg = f"Invalid attempt to instantiate an `Instantiable` directly. "
            raise RuntimeError(msg)

        @classmethod
        def __get_validators__(cls):
            yield assert_instantiable

else:
    from .datatype import BeforeValidator
    from typing import Annotated

    Instantiable = Annotated[InstantiableUnion, BeforeValidator(assert_instantiable)]
    Instantiable.__doc__ = _doc


__all__ = ["Instantiable", "is_instantiable"]



""" 
# Hdl21 Netlisting 
"""

from typing import IO, Union, Optional

# Import the core netlisting from `vlsirtools`
import vlsir
from vlsirtools.netlist import (
    netlist as vlsir_netlist,
    NetlistFormat,
    NetlistFormatSpec,
    NetlistOptions,
)

from .elab import Elaboratables
from .proto import to_proto


def netlist(
    src: Union[Elaboratables, vlsir.circuit.Package],
    dest: IO,
    *,
    domain: Optional[str] = None,
    **kwargs
) -> None:
    """
    # Hdl21 Netlisting

    Netlist one or more Hdl21 `Elaboratable`s - typically `Module`s - or a VLSIR `Package` to destination `dest`.
    All other options are forwarded to the underlying VLSIR netlister.

    Example usages:
    ```python
    h.netlist(MyModule, dest=open('mynetlist.v', 'w'), fmt='verilog')
    ```
    ```python
    s = StringIO()
    h.netlist(MyGenerator(Params()), dest=s, fmt='spectre')
    ```
    ```python
    import sys
    h.netlist(h.to_proto([MyModule, MyGenerator(Params())]), dest=sys.stdout, fmt='spice')
    ```
    """

    # Convert to the Vlsir `Package` if necessary
    if not isinstance(src, vlsir.circuit.Package):
        pkg = to_proto(top=src, domain=domain)
    else:
        pkg = src

    # And invoke the VLSIR netlister
    return vlsir_netlist(pkg=pkg, dest=dest, **kwargs)


__all__ = ["netlist", "NetlistFormat", "NetlistFormatSpec", "NetlistOptions"]
